# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'pcassie'
copyright = '2025, Kenny T. Phan'
author = 'Kenny T. Phan'
release = '0.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.todo",
    "sphinx.ext.napoleon", 
    "sphinx.ext.viewcode", 
    "nbsphinx", 
    "sphinx.ext.autodoc"]

templates_path = ['_templates']
exclude_patterns = ['.ipynb_checkpoints/*']

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'bizstyle'
html_static_path = ['_static']

# additions
nbsphinx_prompt_width = 0 # no prompts in nbsphinx
master_doc = 'index'